from abc import abstractmethod, ABCMeta
from collections import OrderedDict
from copy import deepcopy
from logging import Logger
from typing import List, Text, Tuple, Union, Dict, Any

from waterch.tasker.mixin import value
from waterch.tasker.storage import Storage
from waterch.tasker.tasks.containers import Task
from waterch.tasker.typedef import Definition, Profile
from waterch.tasker.typedef import Return
from waterch.tasker.utils import import_reference

try:
    import torch
    from torch import nn
    from torch import optim
    from torch.utils import data
    from ignite import engine
    from ignite import metrics
except ImportError as ie:
    raise RuntimeError('Tasks in this module needs pytorch and pytorch-ignite modules')


class TrainTask(Task, metaclass=ABCMeta):
    """
    <b>waterch.tasker.tasks.torch.TrainTask</b>

    The fundamental task construction to train a PyTorch model by provided data loaders.

    You need to run a task providing two data loaders named "train_loader" and "validate_loader"
    in shared storage before this task as well as the trained model will be stored into "model"
    label in shared storage.

    However, many actions should be redefined by users when implementing `TrainTask`. You can
    also implement [`SimpleTrainTask`][waterch.tasker.tasks.torch.SimpleTrainTask] to boost
    your development.
    """

    def __init__(self, prefix: Text = None):
        if prefix is not None:
            self.PROVIDE_KEY = f'{prefix}_model'
        else:
            self.PROVIDE_KEY = 'model'

    def invoke(self, profile: Profile, shared: Storage, logger: Logger) -> int:
        train_loader = shared['train_loader']
        validate_loader = shared['validate_loader']

        model = self.create_model(profile.model, shared, logger)
        optimizer_return = self.create_optimizer(model, profile.optimizer, shared, logger)
        if isinstance(optimizer_return, tuple):
            optimizer = optimizer_return[0]
            scheduler = optimizer_return[1] if len(optimizer_return) > 1 else None
        else:
            optimizer = optimizer_return
            scheduler = None
        loss = self.create_loss(profile.loss, shared, logger)
        metrics_dict = OrderedDict()
        self.more_metrics(metrics_dict)

        def prepare_batch(batch, device=None, non_blocking=False):
            return self.prepare_batch(deepcopy(batch), device, non_blocking)

        trainer = self.create_trainer(
            model, optimizer, loss,
            device=torch.device(profile.device),
            non_blocking=profile.non_blocking,
            prepare_batch=prepare_batch,
        )

        evaluator = self.create_evaluator(
            model, metrics_dict,
            device=torch.device(profile.device),
            non_blocking=profile.non_blocking,
            prepare_batch=prepare_batch,
        )

        @trainer.on(engine.Events.STARTED)
        def on_epoch_started(engine_):
            return self.on_epoch_started(engine_, profile, shared, logger)

        @trainer.on(engine.Events.COMPLETED)
        def on_completed(engine_):
            return self.on_completed(engine_, profile, shared, logger)

        @trainer.on(engine.Events.ITERATION_STARTED)
        def on_iteration_started(engine_):
            return self.on_iteration_started(engine_, profile, shared, logger)

        @trainer.on(engine.Events.ITERATION_COMPLETED)
        def on_iteration_completed(engine_):
            return self.on_iteration_completed(engine_, profile, shared, logger)

        @trainer.on(engine.Events.EPOCH_STARTED)
        def on_epoch_started(engine_):
            return self.on_epoch_started(engine_, profile, shared, logger)

        @trainer.on(engine.Events.EPOCH_COMPLETED)
        def on_epoch_completed(engine_):
            evaluator.run(validate_loader)
            if scheduler is not None:
                scheduler.step(engine_.state.output)
            return self.on_epoch_completed(engine_, evaluator.state.metrics, profile, shared, logger)

        trainer.run(
            train_loader,
            max_epochs=profile.max_epochs
        )
        shared[self.PROVIDE_KEY] = model
        return Return.SUCCESS.value

    def require(self) -> List[Text]:
        """
        The task requires 2 items, including "train_loader" and "validate_loader.

        Returns:
            "train_loader" and "validate_loader
        """
        return ['train_loader', 'validate_loader']

    def provide(self) -> List[Text]:
        """
        The task provides 1 item, including "model".

        Returns:
            "model"
        """
        return [self.PROVIDE_KEY]

    def remove(self) -> List[Text]:
        """
        This task removes nothing.

        Returns:
            Nothing
        """
        return []

    @classmethod
    def define(cls) -> List[Definition]:
        """
        ```toml
        __schema__ = "conv_train.ConvTrainTask"
        device = ""
        non_blocking = true
        max_epochs = 0
        [model]
        # You must implement define_optimizer to define the profile of the model.

        [optimizer]
        # You must implement define_optimizer to define the profile of the optimizer.

        [loss]
        # You must implement define_optimizer to define the profile of the loss function.
        ```
        """
        return [
            value('model', list, cls.define_model()),
            value('optimizer', list, cls.define_optimizer()),
            value('loss', list, cls.define_loss()),
            value('device', str),
            value('non_blocking', bool),
            value('max_epochs', int),
        ]

    @abstractmethod
    def create_model(self, profile: Profile, shared: Storage, logger: Logger) -> nn.Module:
        """
        The function to create model instance with defined profile.

        Args:
            profile: Runtime profile defined in TOML file of model.
            shared: Shared storage in the whole lifecycle.
            logger: The logger named with this Task.

        Returns:
            A new model instance to train in this task.
        """
        raise NotImplementedError(f'Please create the model in create_model')

    @classmethod
    @abstractmethod
    def define_model(cls):
        """
        A profile template of model need to be implemented by user.

        Returns:
            Definition of model profile.
        """
        raise NotImplementedError(f'Please define the model profile in define_model')

    @abstractmethod
    def create_optimizer(self, model: nn.Module, profile: Profile, shared: Storage, logger: Logger) -> Union[
        optim.Optimizer, Tuple[optim.Optimizer, optim.lr_scheduler._LRScheduler]
    ]:
        """
        The function to create optimizer instance with defined profile.

        Args:
            model: The model need to be optimized.
            profile: Runtime profile defined in TOML file of optimizer.
            shared: Shared storage in the whole lifecycle.
            logger: The logger named with this Task.

        Returns:
            A new optimizer instance to train in this task. A optional scheduler can also be returned with
            the optimizer.
        """
        raise NotImplementedError(f'Please create the optimizer (or with a lr_scheduler) in create_optimizer')

    @classmethod
    @abstractmethod
    def define_optimizer(cls):
        """
        A profile template of optimizer need to be implemented by user.

        Returns:
            Definition of optimizer profile.
        """
        raise NotImplementedError(f'Please define the optimizer profile in define_optimizer')

    @abstractmethod
    def create_loss(self, profile: Profile, shared: Storage, logger: Logger) -> nn.Module:
        """
        The function to create loss function instance with defined profile.

        Args:
            profile: Runtime profile defined in TOML file of model.
            shared: Shared storage in the whole lifecycle.
            logger: The logger named with this Task.

        Returns:
            A new loss function instance to train in this task.
        """
        raise NotImplementedError(f'Please create the loss function in create_loss')

    @classmethod
    @abstractmethod
    def define_loss(cls):
        """
        A profile template of loss function need to be implemented by user.

        Returns:
            Definition of loss function profile.
        """
        raise NotImplementedError(f'Please define the loss function profile in define_loss')

    def more_metrics(self, metrics_: OrderedDict):
        """
        All metric method you want to monitor in training epochs.

        Args:
            metrics_: The metrics collections.

        Returns:
            Nothing
        """
        pass

    def prepare_batch(self, batch, device=None, non_blocking=False):
        """
        The method to prepare batch provided by data loader. The value returned by this method
        should be a tuple with the format such as (x, y).

        Args:
            batch: Raw batch produced by data loader
            device: The device which model should be moved to
            non_blocking: If True and this copy is between CPU and GPU, the copy may occur asynchronously with respect to the host. For other cases, this argument has no effect.

        Returns:
            The output tuple with format (x, y)

        """
        x, y = batch
        return (
            engine.convert_tensor(x, device=device, non_blocking=non_blocking),
            engine.convert_tensor(y, device=device, non_blocking=non_blocking)
        )

    def on_epoch_started(self, engine_: engine.Engine, profile: Profile, shared: Storage, logger: Logger):
        """
        Define the actions when an epoch started.

        Args:
            engine_: Training context
            profile: The profile of task
            shared: Shared storage of task
            logger: Logger of task.

        Returns:
            Nothing
        """
        pass

    def on_epoch_completed(self, engine_: engine.Engine, metrics_: Dict[Text, Any], profile: Profile, shared: Storage,
                           logger: Logger):
        """
        Define the actions when an epoch completed.

        Args:
            engine_: Training context
            metrics_: Calculated metric values of validate dataset
            profile: The profile of task
            shared: Shared storage of task
            logger: Logger of task.

        Returns:
            Nothing
        """
        pass

    def on_started(self, engine_: engine.Engine, profile: Profile, shared: Storage, logger: Logger):
        """
        Define the actions when the task completed.

        Args:
            engine_: Training context
            profile: The profile of task
            shared: Shared storage of task
            logger: Logger of task.

        Returns:
            Nothing
        """
        pass

    def on_completed(self, engine_: engine.Engine, profile: Profile, shared: Storage, logger: Logger):
        """
        Define the actions when the task completed.

        Args:
            engine_: Training context
            profile: The profile of task
            shared: Shared storage of task
            logger: Logger of task.

        Returns:
            Nothing
        """
        pass

    def on_iteration_started(self, engine_: engine.Engine, profile: Profile, shared: Storage, logger: Logger):
        """
        Define the actions when an iteration started.

        Args:
            engine_: Training context
            profile: The profile of task
            shared: Shared storage of task
            logger: Logger of task.

        Returns:
            Nothing
        """
        pass

    def on_iteration_completed(self, engine_: engine.Engine, profile: Profile, shared: Storage, logger: Logger):
        """
        Define the actions when an iteration completed.

        Args:
            engine_: Training context
            profile: The profile of task
            shared: Shared storage of task
            logger: Logger of task.

        Returns:
            Nothing
        """
        pass

    def create_trainer(
            self, model, optimizer, loss_fn, device, non_blocking, prepare_batch, output_transform=lambda x, y, y_pred, loss: loss.item()
    ):
        return engine.create_supervised_trainer(
            model, optimizer, loss_fn, device, non_blocking, prepare_batch, output_transform
        )

    def create_evaluator(
            self, model, metrics, device, non_blocking, prepare_batch, output_transform=lambda x, y, y_pred: (y_pred, y,)
    ):
        return engine.create_supervised_evaluator(
            model, metrics, device, non_blocking, prepare_batch, output_transform
        )


class SimpleTrainTask(TrainTask, metaclass=ABCMeta):
    """
    <b>waterch.tasker.tasks.torch.SimpleTrainTask</b>

    An easy to use base class of task for training models. You don't need to modify the code
    to create optimizer and loss function, instead, you only need to implement the model.
    """

    def create_optimizer(self, model: nn.Module, profile: Profile, shared: Storage, logger: Logger) -> Union[
        optim.Optimizer, Tuple[optim.Optimizer, optim.lr_scheduler._LRScheduler]
    ]:
        """
        Automatically create optimizer from profile.

        ```toml
        [optimizer]
        reference = ""
            [optimizer.kwargs]
            # the fields in kwargs should be filled according to the definition of optimizer class.
            [optimizer.scheduler] # optional
            reference = ""
                [optimizer.scheduler.kwargs]
                # the fields in kwargs should be filled according to the definition of scheduler class.
        ```

        Args:
            model: The model need to be optimized.
            profile: Runtime profile defined in TOML file of optimizer.
            shared: Shared storage in the whole lifecycle.
            logger: The logger named with this Task.

        Returns:
            A new optimizer instance to train in this task. A optional scheduler can also be returned with
            the optimizer.
        """
        optimizer_class = import_reference(profile.reference)
        optimizer = optimizer_class(params=model.parameters(), **profile.kwargs)
        if 'scheduler' in profile:
            scheduler_class = import_reference(profile.scheduler.reference)
            scheduler = scheduler_class(optimizer=optimizer, **profile.scheduler.kwargs)
            return optimizer, scheduler
        else:
            return optimizer

    @classmethod
    def define_optimizer(cls):
        return [
            value('reference', str),
            value('kwargs', list),
            value('scheduler', list, [
                value('reference', str),
                value('kwargs', list),
                value('# OPTIONAL', str)
            ])
        ]

    def create_loss(self, profile: Profile, shared: Storage, logger: Logger) -> nn.Module:
        """
        Automatically create loss function from profile.

        Args:
            profile: Runtime profile defined in TOML file of loss function.
            shared: Shared storage in the whole lifecycle.
            logger: The logger named with this Task.

        Returns:
            A new loss function instance to train in this task.
        """
        loss_class = import_reference(profile.reference)
        loss = loss_class(**profile.kwargs)
        return loss

    @classmethod
    def define_loss(cls):
        return [
            value('reference', str),
            value('kwargs', list)
        ]

    def on_epoch_completed(self,
                           engine_: engine.Engine, metrics_: Dict[Text, Any], profile: Profile, shared: Storage,
                           logger: Logger):
        """
        Display metrics output of each epochs which declared in `more_metrics`.
        """
        logger.info(f'EPOCH {engine_.state.epoch} | ' + ' | '.join(
            map(
                lambda it: f'{it[0]}: {it[1]}',
                metrics_.items()
            )
        ))

    def on_iteration_completed(self, engine_: engine.Engine, profile: Profile, shared: Storage, logger: Logger):
        """
        Display loss output of each iterations.
        """
        epoch_iteration = engine_.state.iteration % engine_.state.epoch_length
        if epoch_iteration == 0:
            epoch_iteration = engine_.state.epoch_length
        logger.debug(f'ITERATION {epoch_iteration} | output: {engine_.state.output}')


class DataLoaderTask(Task):
    """
    <b>waterch.tasker.tasks.torch.DataLoaderTask</b>

    The fundamental task construction to provide data loaders.

    Please declare the prefix of data loader in shared storage in reference
    field of meta definitions.

    Examples:
        ```toml
        # Data loader of validation.
        [[__meta__]]
        reference = "cifar_loader.CIFARDataLoaderTask:validate"
        include = false
        path = ""
        profile = "validate_loader"
        execute = true

        # Data Loader of training.
        [[__meta__]]
        reference = "conv_train.ConvTrainTask:train"
        include = false
        path = ""
        profile = "train"
        execute = true
        ```
    """

    def __init__(self, _type: Text = None):
        if _type is None:
            self.PROVIDE_KEY = 'loader'
        else:
            self.PROVIDE_KEY = f'{_type}_loader'

    def invoke(self, profile: Profile, shared: Storage, logger: Logger) -> int:
        dataset = self.create_dataset(profile.dataset, shared, logger)

        assert isinstance(dataset, data.Dataset)

        loader_params = dict(profile.loader)
        loader_params['dataset'] = dataset

        assert profile.sampler_type in ('sampler', 'batch_sampler', 'none')

        if profile.sampler_type != 'none':
            loader_params[profile.sampler_type] = self.create_sampler(
                dataset, profile.sampler_type == 'batch_sampler', profile.sampler,
                shared, logger
            )

        shared[self.PROVIDE_KEY] = data.DataLoader(**loader_params)
        return Return.SUCCESS.value

    def require(self) -> List[Text]:
        return []

    def provide(self) -> List[Text]:
        return [self.PROVIDE_KEY]

    def remove(self) -> List[Text]:
        return []

    @classmethod
    def define(cls) -> List[Definition]:
        """
        ```toml
        __schema__ = "waterch.tasker.tasks.torch.DataLoaderTask"
        sampler_type = ""

        [loader]
        batch_size = 0
        shuffle = true
        num_workers = 0
        pin_memory = true
        drop_last = true
        ```
        """
        return [
            value('dataset', list, cls.define_dataset()),
            value('loader', list, [
                value('batch_size', int),
                value('shuffle', bool),
                value('num_workers', int),
                value('pin_memory', bool),
                value('drop_last', bool),
            ]),
            value('sampler_type', str),
            value('sampler', list, cls.define_sampler())
        ]

    @abstractmethod
    def create_dataset(self, profile: Profile, shared: Storage, logger: Logger) -> data.Dataset:
        """
        The function to create dataset instance with defined profile.

        Args:
            profile: Runtime profile defined in TOML file of dataset.
            shared: Shared storage in the whole lifecycle.
            logger: The logger named with this Task.

        Returns:
            A new dataset instance.
        """
        raise NotImplementedError('Please create the dataset in create_dataset')

    @classmethod
    @abstractmethod
    def define_dataset(cls):
        """
        A profile template of dataset need to be implemented by user.

        Returns:
            Definition of dataset profile.
        """
        raise NotImplementedError('Please define the dataset profile in define_dataset')

    @abstractmethod
    def create_sampler(self, dataset: data.Dataset, batch_sampler: bool, profile: Profile, shared: Storage,
                       logger: Logger):
        """
        The function to create sampler instance with defined profile.

        Args:
            dataset: The dataset instance need to be loaded.
            batch_sampler: Whether to use batch_sampler.
            profile: Runtime profile defined in TOML file of sampler or batch sampler.
            shared: Shared storage in the whole lifecycle.
            logger: The logger named with this Task.

        Returns:
            A new sampler or batch sampler instance.
        """
        if batch_sampler:
            raise NotImplementedError('Please create the batch sampler in create_sampler')
        else:
            raise NotImplementedError('Please create the sampler in create_sampler')

    @classmethod
    @abstractmethod
    def define_sampler(cls):
        raise NotImplementedError('Please define the sampler or batch sampler profile in define_sampler')


class SimpleDataLoaderTask(DataLoaderTask):
    """
    <b>waterch.tasker.tasks.torch.SimpleDataLoaderTask</b>

    An easy to use base class of task for providing data loader. You can
    create data loader only with reference of dataset and related profile.
    """

    def create_dataset(self, profile: Profile, shared: Storage, logger: Logger) -> data.Dataset:
        dataset_cls = import_reference(profile.reference)
        return dataset_cls(**profile.kwargs)

    @classmethod
    def define_dataset(cls):
        return [
            value('reference', str),
            value('kwargs', list, [])
        ]

    def create_sampler(self, dataset: data.Dataset, batch_sampler: bool, profile: Profile, shared: Storage,
                       logger: Logger):
        return None

    @classmethod
    def define_sampler(cls):
        return []



__all__ = [
    'Profile',
    'Definition',
    'Return',
    'storage',
    'mixin',
    'value',
    'include',
    'launch',
    'utils',
    'tasks',
    'version'
]

from waterch.tasker import storage, mixin, utils, tasks
from waterch.tasker.launcher import launch
from waterch.tasker.mixin import value, include
from waterch.tasker.typedef import Profile, Definition, Return
from waterch.tasker._version import version



from datetime import datetime
from logging import Logger
from os import environ
from typing import List, Text
from uuid import uuid4

from waterch.tasker.typedef import Definition, Profile, Return
from waterch.tasker.mixin import value
from waterch.tasker.storage import Storage
from waterch.tasker.tasks.containers import Task


class GenerateIdTask(Task):
    """
    <b>waterch.tasker.tasks.utils.GenerateIdTask</b>

    Task to generate ID.
    The task will declare a field named `id` in shared storage.

    Currently, `GenerateIdTask` supports 3 type of items to generate ID,
    including "datetime", "uuid", "label". If you need to label something,
    considerate using "label" into "join" fields. Items will be joined by
    string defined in field "by".
    """

    def generate_datetime(self, profile) -> Text:
        return datetime.now().strftime('%Y%m%d%H%M%S')

    def generate_uuid(self, profile) -> Text:
        return uuid4().hex.replace('-', '')

    def generate_label(self, profile) -> Text:
        try:
            return profile.label
        except Exception:
            raise RuntimeError('Please define "label" attribute in profile when including label item.')

    def invoke(self, profile: Profile, shared: Storage, logger: Logger) -> Return:
        """
        Generate ID.
        Args:
            profile: Runtime profile defined in TOML file.
            shared: Shared storage in the whole lifecycle.
            logger: The logger named with this Task.

        Returns:
            Always return [SUCCESS][Return.SUCCESS].
        """

        def generate(name: Text) -> Text:
            generator = getattr(self, f'generate_{name}', None)
            if generator is None:
                raise RuntimeError(f'ID item {name} is invalid, please check and fix.')
            return generator(profile)

        shared['id'] = profile.by.join(filter(lambda t: t != '', map(generate, profile.join)))
        logger.debug(f'ID: {shared["id"]}')
        return Return.SUCCESS.value

    def require(self) -> List[Text]:
        """
        This task requires nothing.
        Returns:
            Nothing
        """
        return ['id']

    def provide(self) -> List[Text]:
        """
        This task provides:

        | Key  | Value                                 |
        | ---- | ------------------------------------- |
        | id   | Generated ID provided to other tasks. |

        Returns:
            Only contains "id"
        """
        return ['id']

    def remove(self) -> List[Text]:
        """
        This task removes nothing.
        Returns:
            Nothing
        """

    @classmethod
    def define(cls) -> List[Definition]:
        """
        Examples:
            ```toml
            __schema__ = "waterch.tasker.tasks.utils.GenerateIdTask"
            by = ""
            label = ""
            ```
        Returns:
            Schema of profile
        """
        return [
            value('by', str),
            value('join', list, [str]),
            value('label', str)
        ]


class SetEnvironmentTask(Task):
    """
    <b>waterch.tasker.tasks.utils.SetEnvironmentTask</b>

    Task to set environment values.
    """

    def invoke(self, profile: Profile, shared: Storage, logger: Logger) -> int:
        """
        Set environment values.
        Args:
            profile: Runtime profile defined in TOML file.
            shared: Shared storage in the whole lifecycle.
            logger: The logger named with this Task.

        Returns:
            Always return [SUCCESS][Return.SUCCESS].
        """
        for item in profile.env:
            assert len(item) == 2
            environ[item[0]] = item[1]
        logger.debug(f'Environment:')
        for item in environ.items():
            logger.debug(f'  {"=".join(item)}')
        return Return.SUCCESS.value

    def require(self) -> List[Text]:
        """
        This task requires nothing.

        Returns:
            Nothing
        """
        return []

    def provide(self) -> List[Text]:
        """
        This task provides nothing.

        Returns:
            Nothing
        """
        return []

    def remove(self) -> List[Text]:
        """
        This task removes nothing.

        Returns:
            Nothing
        """
        return []

    @classmethod
    def define(cls) -> List[Definition]:
        """
        ```toml
        __schema__ = "waterch.tasker.tasks.utils.SetEnvironmentTask"
        env = [
            ["CUDA_VISIBLE_DEVICES", "0"],
        ]
        ```

        Returns:
            Schema of profile.
        """
        return [
            value('env', list, [[str]]),
        ]





__all__ = [
    'Storage',
    'DictStorage',
    'StorageView',
    'CommonStorageView',
    'ForkStorageView'
]

from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from sys import stderr
from typing import Text, Any


class Storage(metaclass=ABCMeta):
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def __getitem__(self, item):
        raise NotImplementedError('Please implement __getitem__ function instead of using abstract storage.')

    @abstractmethod
    def __setitem__(self, key, value):
        raise NotImplementedError('Please implement __setitem__ function instead of using abstract storage.')

    @abstractmethod
    def __delitem__(self, key):
        raise NotImplementedError('Please implement __delitem__ function instead of using abstract storage.')

    @abstractmethod
    def __contains__(self, item):
        raise NotImplementedError('Please implement __contains__ function instead of using abstract storage.')

    @abstractmethod
    def load(self):
        raise NotImplementedError('Please implement load function instead of using abstract storage.')

    @abstractmethod
    def dump(self):
        raise NotImplementedError('Please implement dump function instead of using abstract storage.')


class DictStorage(Storage):
    __slots__ = ['_data']

    def __init__(self, **kwargs):
        super(DictStorage, self).__init__(**kwargs)
        self._data = OrderedDict()

    def __getitem__(self, item: Text) -> Any:
        return self._data.__getitem__(item)

    def __setitem__(self, key: Text, value: Any):
        return self._data.__setitem__(key, value)

    def __delitem__(self, key: Text):
        return self._data.__delitem__(key)

    def __contains__(self, item):
        return self._data.__contains__(item)

    def load(self):
        print('Function load is invalid in DictStorage.', file=stderr)

    def dump(self):
        print('Function dump is invalid in DictStorage.', file=stderr)


class StorageView(Storage):
    @abstractmethod
    def storage(self) -> Storage:
        raise RuntimeError(f'Please implement storage function instead of using abstract storage view.')

    @staticmethod
    def require_check(task, item):
        if item not in task.require() and not item.startswith('__'):
            raise RuntimeError(f'Forbidden reading on key "{item}", please declare it in method require.')

    @staticmethod
    def provide_check(task, item):
        if item not in task.provide() and not item.startswith('__'):
            raise RuntimeError(f'Forbidden writing on key "{item}", please declare it in method provide.')

    @staticmethod
    def remove_check(task, item):
        if item not in task.remove() and not item.startswith('__'):
            raise RuntimeError(f'Forbidden deleting on key "{item}", please declare it in method remove.')


class CommonStorageView(StorageView):
    def __init__(self, **kwargs):
        super(CommonStorageView, self).__init__(**kwargs)
        assert 'storage' in kwargs
        assert 'task' in kwargs
        self._task = kwargs['task']
        self._storage = kwargs['storage']

    def storage(self) -> Storage:
        return self._storage

    def __getitem__(self, item):
        self.require_check(self._task, item)
        return self._storage.__getitem__(item)

    def __setitem__(self, key, value):
        self.provide_check(self._task, key)
        return self._storage.__setitem__(key, value)

    def __delitem__(self, key):
        self.remove_check(self._task, key)
        return self._storage.__delitem__(key)

    def __contains__(self, item):
        return self._storage.__contains__(item)

    def load(self):
        return self._storage.load()

    def dump(self):
        return self._storage.dump()


class ForkStorageView(StorageView):
    META_KEY = '__tasks__'

    def __init__(self, **kwargs):
        super(ForkStorageView, self).__init__(**kwargs)
        assert 'storage' in kwargs
        assert 'task' in kwargs

        self._storage: Storage = kwargs['storage']
        self._task = kwargs['task']
        if self.META_KEY in self._storage:
            self._storage[self.META_KEY].append(hex(hash(self._task)))
        else:
            self._storage[self.META_KEY] = [hex(hash(self._task))]

    def _mirror_key(self, key) -> Text:
        return f'{key}@{hex(hash(self._task))}'

    def __getitem__(self, item):
        self.require_check(self._task, item)
        if item in self._storage:
            return self._storage.__getitem__(item)
        else:
            return self._storage.__getitem__(self._mirror_key(item))

    def __setitem__(self, key, value):
        self.provide_check(self._task, key)
        if key in self._storage:
            return self._storage.__setitem__(key, value)
        else:
            return self._storage.__setitem__(self._mirror_key(key), value)

    def __delitem__(self, key):
        self.remove_check(self._task, key)
        if key in self._storage:
            return self._storage.__delitem__(key)
        else:
            return self._storage.__delitem__(self._mirror_key(key))

    def __contains__(self, item):
        return self._storage.__contains__(item) or self._storage.__contains__(self._mirror_key(item))

    def load(self):
        return self._storage.load()

    def dump(self):
        return self._storage.dump()

    def storage(self) -> Storage:
        return self._storage


class MapStorageView(StorageView):
    META_KEY = ForkStorageView.META_KEY
    MIRROR_KEY = '__mirrors__'

    def __init__(self, **kwargs):
        super(MapStorageView, self).__init__(**kwargs)
        self._storage = kwargs['storage']
        self._task = kwargs['task']
        self._mirror = kwargs['mirror']

        if self.META_KEY not in self._storage:
            raise RuntimeError('Please run a ForkTask instance before executing a MapTask')
        elif self._mirror not in self._storage[self.META_KEY]:
            raise RuntimeError('Failed to access forked keys in shared storage')

        if self.MIRROR_KEY not in self._storage:
            self._storage[self.MIRROR_KEY] = {}
        if self._mirror not in self._storage[self.MIRROR_KEY]:
            self._storage[self.MIRROR_KEY][self._mirror] = [hex(hash(self._task))]
        else:
            self._storage[self.MIRROR_KEY][self._mirror].append(hex(hash(self._task)))

    def storage(self) -> Storage:
        return self._storage

    def __getitem__(self, item):
        self.require_check(self._task, item)
        if item in self._storage:
            return self._storage.__getitem__(item)
        elif f'{item}@{self._mirror}' in self._storage:
            return self._storage.__getitem__(f'{item}@{self._mirror}')
        else:
            for mirror in reversed(self._storage[self.MIRROR_KEY]):
                if f'{item}@{mirror}' in self._storage:
                    return self._storage.__getitem__(f'{item}@{mirror}')
            raise KeyError(f'Key {item} is not found in storage.')

    def __setitem__(self, key, value):
        self.provide_check(self._task, key)
        if key in self._storage:
            return self._storage.__setitem__(key, value)
        elif f'{key}@{self._mirror}' in self._storage:
            del self._storage[f'{key}@{self._mirror}']
            return self._storage.__setitem__(f'{key}@{self._mirror}', value)
        else:
            for mirror in self._storage[self.MIRROR_KEY]:
                if f'{key}@{mirror}' in self._storage:
                    del self._storage[f'{key}@{mirror}']
                    return self._storage.__setitem__(f'{key}@{mirror}', value)
        return self._storage.__setitem__(f'{key}@{hex(hash(self._task))}', value)

    def __delitem__(self, key):
        self.remove_check(self._task, key)
        if key in self._storage:
            self._storage.__delitem__(key)
        if f'{key}@{self._mirror}' in self._storage:
            self._storage.__delitem__(f'{key}@{self._mirror}')
        for mirror in self._storage[self.MIRROR_KEY]:
            if f'{key}@{mirror}' in self._storage:
                self._storage.__delitem__(f'{key}@{mirror}')
        raise KeyError(f'Key {key} is not found in storage.')

    def __contains__(self, item):
        if item in self._storage:
            return self._storage.__contains__(f'{item}@{self._mirror}')
        if f'{item}@{self._mirror}' in self._storage:
            return self._storage.__contains__(f'{item}@{self._mirror}')
        for mirror in self._storage[self.MIRROR_KEY]:
            if f'{item}@{mirror}' in self._storage:
                return self._storage.__contains__(f'{item}@{mirror}')
        return False

    def load(self):
        return self._storage.load()

    def dump(self):
        return self._storage.dump()


class ReduceStorageView(MapStorageView):
    def __setitem__(self, key, value):
        return self._storage.__setitem__(key, value)

    def __delitem__(self, key):
        raise RuntimeError('Delete operation is forbidden in reduce task.')


__all__ = [
    'Launcher',
    'launch'
]

from argparse import ArgumentParser, Namespace
from datetime import datetime
from logging import getLogger as get_logger, basicConfig as basic_config
from os import makedirs, path, curdir
from sys import stdout, stderr, path as syspath
from typing import List

from toml import dump as toml_dump

from waterch.tasker.mixin import ProfileMixin, value
from waterch.tasker.storage import DictStorage, CommonStorageView
from waterch.tasker.tasks import Task
from waterch.tasker.typedef import Profile, Return, Definition
from waterch.tasker.utils import import_reference, extract_reference
from waterch.tasker._version import version


class Launcher(ProfileMixin):
    @classmethod
    def define(cls) -> List[Definition]:
        return [
            value('__name__', str),
            value('__author__', str),
            value('__version__', str),
            value('__email__', str),
            value('__abstract__', str),
            value('__setting__', list, [
                value('storage', list, [
                    value('reference', str)
                ]),
                value('log', list, [
                    value('stdout', bool),
                    value('level', str),
                ])
            ]),
            value('__meta__', list, [
                [
                    value('reference', str),
                    value('include', bool),
                    value('path', str),
                    value('profile', str),
                    value('execute', bool)
                ]
            ]),
        ]

    def _argparser(self) -> ArgumentParser:
        parser = ArgumentParser(
            prog='waterch-tasker',
            description='A scalable and extendable experiment task scheduler framework'
        )
        parser.add_argument(
            '-v', '--version',
            action='version',
            version=f'%(prog)s {version}',
        )
        subcommands = parser.add_subparsers(
            title='command',
            description='execute a command of waterch-tasker',
            dest='command',
            help='choose a command',
        )
        launch_parser = subcommands.add_parser(
            'launch',
            help='launch tasks by a profile'
        )
        launch_parser.add_argument(
            '-f', '--file',
            nargs=1,
            required=True,
            help='launch tasks defined by this file'
        )
        template_parser = subcommands.add_parser(
            'template',
            help='generate a profile template by a class reference',
        )
        template_parser.add_argument(
            '-r', '--reference',
            nargs='+',
            required=True,
            help='generate profile template of those tasks'
        )
        template_parser.add_argument(
            '-o', '--output',
            nargs=1,
            required=False,
            help='output folder path',
            default=['-']
        )
        return parser

    def argparse(self) -> Namespace:
        return self._argparser().parse_args()

    def invoke(self, namespace: Namespace):
        syspath.append(curdir)
        command = getattr(self, f'command_{namespace.command}', self.command_unknown)
        command(namespace)

    def command_unknown(self, namespace: Namespace):
        print(
            f'Please check the input arguments,\n'
            f'command "{namespace.command}" not exists.'
        )

    def command_template(self, namespace: Namespace):
        slash_number = 20
        enable_stdout = namespace.output[0] == '-'
        if not enable_stdout and not path.exists(namespace.output[0]):
            makedirs(namespace.output[0])

        for reference in namespace.reference:
            stream = stdout if enable_stdout else open(path.join(namespace.output[0], f'{reference}.toml'), 'w')
            if enable_stdout:
                print(f'Profile of {reference}')
                print('-' * slash_number)
            cls = import_reference(reference)
            if not issubclass(cls, ProfileMixin):
                raise RuntimeError(f'Class {cls} do NOT supported profile.')
            if not issubclass(cls, Task):
                print(f'Class {cls} is NOT a Task, this may cause unpredictable issue.', file=stderr)
            toml_dump(cls.profile_template(), stream)
            if not enable_stdout:
                stream.flush()
                stream.close()
            else:
                print('-' * slash_number)

    def command_launch(self, namespace):
        slash_number = 20
        profile = Profile.from_toml(filename=namespace.file[0])
        # Check the profile
        missing_key = list(
            map(
                lambda definition: definition.name,
                filter(
                    lambda value: value.name not in profile,
                    self.define()
                ))
        )
        if len(missing_key) > 0:
            raise RuntimeError(f'Key {",".join(missing_key)} is/are missing.')
        # Print header info of profile
        print('-' * slash_number)
        print(f'{profile.__name__} ({profile.__version__})')
        print(f'Author: {profile.__author__}')
        print(f'E-Mail: {profile.__email__}')
        print(profile.__abstract__)
        print('-' * slash_number)
        print()
        # Create shared storage
        if '__setting__' in profile \
                and 'storage' in profile.__setting__ \
                and 'reference' in profile.__setting__.storage:
            try:
                storage_cls = import_reference(profile.__setting__.storage.reference)
            except RuntimeError:
                print('Failed to get storage class, fall back to DictStorage.', file=stderr)
                storage_cls = DictStorage
        else:
            storage_cls = DictStorage
        shared = storage_cls(**profile.__setting__.storage)
        # Configure logging
        log_datetime_format = '%Y-%m-%dT%H:%M:%S'
        log_format = '%(asctime)s|%(levelname)s|%(name)s> %(message)s'
        if profile.__setting__.log.stdout:
            basic_config(
                stream=stdout,
                level=profile.__setting__.log.level,
                format=log_format,
                datefmt=log_datetime_format,
            )
        else:
            basic_config(
                filename=f'logs/{datetime.now().strftime("%Y-%m-%dT%H:%M:%S.log")}',
                format=log_format,
                datefmt=log_datetime_format,
            )
        # Launch tasks
        meta_index = 0
        while meta_index < len(profile.__meta__):
            meta = profile.__meta__[meta_index]
            reference, rparams = extract_reference(meta.reference)
            task_cls = import_reference(reference)
            try:
                if meta.include:
                    task_profile = Profile.from_toml(filename=meta.path)
                else:
                    task_profile = profile[meta.profile]
            except Exception:
                print('Failed to access task profile, stop running.')
                shared.dump()
                break
            task: Task = task_cls(*rparams)
            task_display = f'{reference}[{hex(hash(task))}]'
            task_logger = get_logger(task_display)
            print(task_display)
            print(f'require: {" ".join(task.require())}')
            print(f'provide: {" ".join(task.provide())}')
            print(f'{"-" * (slash_number - 1)}>')
            start_time = datetime.now()
            user_kill = False
            try:
                state = task.invoke(task_profile, CommonStorageView(storage=shared, task=task), task_logger)
            except KeyboardInterrupt:
                state = Return.WRITE.value
                user_kill = True
            end_time = datetime.now()
            print(f'<{"-" * (slash_number - 1)}')
            state_label = 'Failed' if state & Return.ERROR.value else 'Successfully finished'
            print(f'{state_label} in {(end_time - start_time).total_seconds()} seconds.')
            print()
            if state & Return.WRITE.value:
                shared.dump()
            if state & Return.READ.value:
                shared.load()
            if state & Return.EXIT.value:
                print('Stopped by task.')
                break
            if user_kill:
                print('Stopped by user.')
                break
            if not (state & Return.RETRY.value):
                meta_index += 1


def launch():
    launcher = Launcher()
    launcher.invoke(launcher.argparse())
